import os
from typing import List

import soundfile as sf
import torch
from diffusers import DiffusionPipeline
from nltk.tokenize import sent_tokenize
from PIL.Image import Image
from transformers import pipeline
from TTS.api import TTS

from storyteller import StoryTellerConfig
from storyteller.utils import (
    make_timeline_string,
    require_ffmpeg,
    require_punkt,
    subprocess_run,
)


class StoryTeller:
    @require_ffmpeg
    @require_punkt
    def __init__(self, config: StoryTellerConfig):
        self.config = config
        writer_device = torch.device(config.writer_device)
        painter_device = torch.device(config.writer_device)
        self.writer = pipeline(
            "text-generation",
            model=config.writer,
            device=writer_device,
            torch_dtype=getattr(torch, config.writer_dtype),
        )
        self.painter = DiffusionPipeline.from_pretrained(
            config.painter,
            use_auth_token=False,
            torch_dtype=getattr(torch, config.painter_dtype),
        ).to(painter_device)
        if config.enable_attention_slicing:
            self.painter.enable_attention_slicing()
        self.speaker = TTS(config.speaker)
        self.sample_rate = self.speaker.synthesizer.output_sample_rate
        self.output_dir = None

    @classmethod
    def from_default(cls) -> "StoryTeller":
        config = StoryTellerConfig()
        return cls(config)

    @torch.inference_mode()
    def paint(self, prompt: str) -> Image:
        return self.painter(prompt).images[0]

    @torch.inference_mode()
    def speak(self, prompt: str) -> List[int]:
        return self.speaker.tts(prompt)

    @torch.inference_mode()
    def write(self, prompt: str) -> str:
        return self.writer(prompt, max_new_tokens=self.config.max_new_tokens)[0][
            "generated_text"
        ]

    def get_output_path(self, file: str) -> str:
        return os.path.join(self.output_dir, file)

    def generate(
        self,
        writer_prompt: str,
        painter_prompt_prefix: str,
        num_images: int,
        output_dir: str,
    ) -> None:
        video_paths = []
        self.output_dir = output_dir
        sentences = self.write_story(writer_prompt, num_images)
        for i, sentence in enumerate(sentences):
            video_path = self._generate(i, sentence, painter_prompt_prefix)
            video_paths.append(video_path)
        self.concat_videos(video_paths)

    def concat_videos(self, video_paths: List[str]) -> None:
        files_path = self.get_output_path("files.txt")
        output_path = self.get_output_path("out.mp4")
        with open(files_path, "w+") as f:
            for video_path in video_paths:
                f.write(f"file {os.path.split(video_path)[-1]}\n")
        subprocess_run(f"ffmpeg -f concat -i {files_path} -c copy {output_path}")

    def _generate(self, id_: int, sentence: str, painter_prompt_prefix: str) -> str:
        image_path = self.get_output_path(f"{id_}.png")
        audio_path = self.get_output_path(f"{id_}.wav")
        subtitle_path = self.get_output_path(f"{id_}.srt")
        video_path = self.get_output_path(f"{id_}.mp4")
        image = self.paint(f"{painter_prompt_prefix} {sentence}")
        image.save(image_path)
        audio = self.speak(sentence)
        duration, remainder = divmod(len(audio), self.sample_rate)
        if remainder:
            duration += 1
            audio.extend([0] * (self.sample_rate - remainder))
        sf.write(audio_path, audio, self.sample_rate)
        subtitle = f"0\n{make_timeline_string(0, duration)}\n{sentence}"
        with open(subtitle_path, "w+") as f:
            f.write(subtitle)
        subprocess_run(
            f"ffmpeg -loop 1 -i {image_path} -i {audio_path} -vf subtitles={subtitle_path} -tune stillimage -shortest {video_path}"
        )
        return video_path

    def write_story(self, writer_prompt: str, num_sentences: int) -> List[str]:
        sentences = []
        while len(sentences) < num_sentences + 1:
            writer_prompt = self.write(writer_prompt)
            sentences = sent_tokenize(writer_prompt)
        while len(sentences) > num_sentences:
            sentences.pop()
        return sentences



from importlib import metadata

from storyteller.config import StoryTellerConfig
from storyteller.model import StoryTeller

__version__ = metadata.version("storyteller-core")
__all__ = ["__version__", "StoryTellerConfig", "StoryTeller"]



import argparse
import dataclasses
import logging
import os
from dataclasses import dataclass

from storyteller import StoryTeller, StoryTellerConfig
from storyteller.config import StoryTellerConfigArgparseHelpText
from storyteller.utils import set_log_level, set_seed


@dataclass(frozen=True)
class ArgparseDefaults:
    WRITER_PROMPT: str = "Once upon a time, unicorns roamed the Earth."
    PAINTER_PROMPT: str = "Beautiful painting"
    OUTPUT_DIR: str = "out"
    NUM_IMAGES: int = 10
    SEED: int = 42


def get_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--writer_prompt",
        type=str,
        default=ArgparseDefaults.WRITER_PROMPT,
        help=f"The prompt to be used for the writer model. This is the text with which your story will begin. Default: '{ArgparseDefaults.WRITER_PROMPT}'",
    )
    parser.add_argument(
        "--painter_prompt_prefix",
        type=str,
        default=ArgparseDefaults.PAINTER_PROMPT,
        help=f"The prefix to be used for the painter model's prompt. Default: '{ArgparseDefaults.PAINTER_PROMPT}'",
    )
    parser.add_argument(
        "--num_images",
        type=int,
        default=ArgparseDefaults.NUM_IMAGES,
        help=f"The number of images to be generated. Those images will be composed in sequence into a video. Default: {ArgparseDefaults.NUM_IMAGES}",
    )
    parser.add_argument(
        "--output_dir",
        type=str,
        default=ArgparseDefaults.OUTPUT_DIR,
        help=f"The directory to save the generated files to. Default: '{ArgparseDefaults.OUTPUT_DIR}'",
    )
    parser.add_argument(
        "--seed",
        type=int,
        default=ArgparseDefaults.SEED,
        help=f"The seed value to be used for randomization. Default: {ArgparseDefaults.SEED}",
    )
    default_config = StoryTellerConfig()
    for key, value in dataclasses.asdict(default_config).items():
        parser.add_argument(
            f"--{key}",
            type=type(value),
            default=value,
            help=StoryTellerConfigArgparseHelpText.get_help_text_for_var_name(key),
        )
    return parser.parse_args()


def main() -> None:
    args = get_args()
    set_seed(args.seed)
    set_log_level(logging.WARNING)
    os.environ["TOKENIZERS_PARALLELISM"] = "false"
    config = StoryTellerConfig(
        **{
            field.name: getattr(args, field.name)
            for field in dataclasses.fields(StoryTellerConfig)
        }
    )
    story_teller = StoryTeller(config)
    os.makedirs(args.output_dir, exist_ok=True)
    story_teller.generate(
        args.writer_prompt, args.painter_prompt_prefix, args.num_images, args.output_dir
    )


if __name__ == "__main__":
    main()



from __future__ import annotations

import sys
from dataclasses import dataclass
from typing import Dict, Type

import torch


@dataclass(frozen=True)
class StoryTellerConfigDefaults:
    MAX_NEW_TOKENS: int = 50
    WRITER_MODEL: str = "gpt2"
    PAINTER_MODEL: str = "stabilityai/stable-diffusion-2"
    SPEAKER_MODEL: str = "tts_models/en/ljspeech/glow-tts"
    WRITER_DEVICE: str = "cpu"
    PAINTER_DEVICE: str = "cpu"
    WRITER_DTYPE: str = "float32"
    PAINTER_DTYPE: str = "float32"
    ENABLE_ATTENTION_SLICING: bool = False


@dataclass
class StoryTellerConfig:
    max_new_tokens: int = StoryTellerConfigDefaults.MAX_NEW_TOKENS
    writer: str = StoryTellerConfigDefaults.WRITER_MODEL
    painter: str = StoryTellerConfigDefaults.PAINTER_MODEL
    speaker: str = StoryTellerConfigDefaults.SPEAKER_MODEL
    writer_device: str = StoryTellerConfigDefaults.WRITER_DEVICE
    painter_device: str = StoryTellerConfigDefaults.PAINTER_DEVICE
    writer_dtype: str = StoryTellerConfigDefaults.WRITER_DTYPE
    painter_dtype: str = StoryTellerConfigDefaults.PAINTER_DTYPE
    enable_attention_slicing: bool = StoryTellerConfigDefaults.ENABLE_ATTENTION_SLICING

    def __post_init__(self):
        if not hasattr(torch, self.writer_dtype):
            raise ValueError(f"Unsupported torch writer dtype {self.writer_dtype}")
        if not hasattr(torch, self.painter_dtype):
            raise ValueError(f"Unsupported torch painter dtype {self.painter_dtype}")


class StoryTellerConfigArgparseHelpText:
    @staticmethod
    def _get_dataclass_var_name_from_f_string_eq(expression: str) -> str:
        return expression.split(".")[1].split("=")[0]

    # importing from typing for backwards compatibility
    _argparse_help_text: Dict[str, str] = {
        _get_dataclass_var_name_from_f_string_eq(
            f"{StoryTellerConfig.max_new_tokens=}"
        ): f"Maximum number of new tokens to generate in the writer model. Default: {StoryTellerConfigDefaults.MAX_NEW_TOKENS}",
        _get_dataclass_var_name_from_f_string_eq(
            f"{StoryTellerConfig.writer=}"
        ): f"Text generation model to use. Default: '{StoryTellerConfigDefaults.WRITER_MODEL}'",
        _get_dataclass_var_name_from_f_string_eq(
            f"{StoryTellerConfig.painter=}"
        ): f"Image generation model to use. Default: '{StoryTellerConfigDefaults.PAINTER_MODEL}'",
        _get_dataclass_var_name_from_f_string_eq(
            f"{StoryTellerConfig.speaker=}"
        ): f"Text-to-speech (TTS) generation model. Default: '{StoryTellerConfigDefaults.SPEAKER_MODEL}'",
        _get_dataclass_var_name_from_f_string_eq(
            f"{StoryTellerConfig.writer_device=}"
        ): f"Text generation device to use. Default: '{StoryTellerConfigDefaults.WRITER_DEVICE}'",
        _get_dataclass_var_name_from_f_string_eq(
            f"{StoryTellerConfig.painter_device=}"
        ): f"Image generation device to use. Default: '{StoryTellerConfigDefaults.PAINTER_DEVICE}'",
        _get_dataclass_var_name_from_f_string_eq(
            f"{StoryTellerConfig.writer_dtype=}"
        ): f"Text generation dtype to use. Default: '{StoryTellerConfigDefaults.WRITER_DTYPE}'",
        _get_dataclass_var_name_from_f_string_eq(
            f"{StoryTellerConfig.painter_dtype=}"
        ): f"Image generation dtype to use. Default: '{StoryTellerConfigDefaults.PAINTER_DTYPE}'",
        _get_dataclass_var_name_from_f_string_eq(
            f"{StoryTellerConfig.enable_attention_slicing=}"
        ): f"Whether to enable attention slicing for diffusion. Default: '{StoryTellerConfigDefaults.ENABLE_ATTENTION_SLICING}'",
    }

    @classmethod
    def get_help_text_for_var_name(
        cls: Type[StoryTellerConfigArgparseHelpText], var_name: str
    ) -> str:
        try:
            return cls._argparse_help_text[var_name]
        except KeyError:
            print(
                f"Warning: Helper text for {var_name} not found. Returning empty string.",
                file=sys.stderr,
            )
            return ""



import logging
import os
import random
import shutil
import subprocess
from functools import wraps
from typing import Callable

import nltk
import numpy as np
import torch


def require_ffmpeg(func: Callable) -> Callable:
    """Decorator for checking ffmpeg installation."""

    @wraps(func)
    def wrapper_func(*args, **kwargs):
        if shutil.which("ffmpeg") is None:
            raise RuntimeError(
                "`ffmpeg` not found. Please install `ffmpeg` and try again."
            )
        func(*args, **kwargs)

    return wrapper_func


def require_punkt(func: Callable) -> Callable:
    """Decorator for checking nltk punkt module."""

    @wraps(func)
    def wrapper_func(*args, **kwargs):
        try:
            nltk.data.find("tokenizers/punkt")
        except LookupError:
            nltk.download("punkt")
        func(*args, **kwargs)

    return wrapper_func


def make_timeline_string(start: int, end: int) -> str:
    """Create timeline string to write onto .srt subtitle files."""
    start = format_time(start)
    end = format_time(end)
    return f"{start} --> {end}"


def format_time(time: int):
    """Transform time (seconds) to .srt format."""
    mm, ss = divmod(time, 60)
    hh, mm = divmod(mm, 60)
    return f"{hh:02d}:{mm:02d}:{ss:02d},000"


def subprocess_run(command: str) -> None:
    """Wrapper around `subprocess.run()` with /dev/null redirection in stdout and stderr."""
    subprocess.run(
        command.split(), stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
    )


def set_seed(seed: int) -> None:
    """Set seed."""
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = True


def set_log_level(level: int) -> None:
    """Disables specified logging level and below."""
    logging.disable(level)
